@@ -24,8 +24,8 @@
 
 /**
  *  gnet_tcp_socket_connect:
- *  @hostname: Name of host to connect to
- *  @port: Port to connect to
+ *  @hostname: Name of host to connect to.
+ *  @port: Port to connect to.
  *
  *  A quick and easy #GTcpSocket constructor.  This connects to the
  *  specified address and port.  This function does block
@@ -36,7 +36,6 @@
  *  gnet_tcp_socket_get_inetaddr().
  *
  *  Returns: A new #GTcpSocket, or NULL if there was a failure.
- *
  **/
 GTcpSocket*
 gnet_tcp_socket_connect (const gchar* hostname, gint port)
@@ -66,8 +65,8 @@ gnet_tcp_socket_connect (const gchar* hostname, gint port)
  *  connects to the specified address and port and then calls the
  *  callback with the data.  Use this function when you're a client
  *  connecting to a server and you don't want to block or mess with
- *  #GInetAddr's.  It may call the callback before the function
- *  returns.  It will call the callback if there is a failure.
+ *  #GInetAddr's.  It will call the callback if there is a failure.
+ *  It will never call the callback before the function returns.
  *
  *  Returns: ID of the connection which can be used with
  *  gnet_tcp_socket_connect_async_cancel() to cancel it; NULL on
@@ -80,7 +79,6 @@ gnet_tcp_socket_connect_async (const gchar* hostname, gint port,
 			       gpointer data)
 {
   GTcpSocketConnectState* state;
-  gpointer id;
 
   g_return_val_if_fail(hostname != NULL, NULL);
   g_return_val_if_fail(func != NULL, NULL);
@@ -89,18 +87,18 @@ gnet_tcp_socket_connect_async (const gchar* hostname, gint port,
   state->func = func;
   state->data = data;
 
-  id = gnet_inetaddr_new_async (hostname, port,  
-				gnet_tcp_socket_connect_inetaddr_cb, 
-				state);
+  state->inetaddr_id = 
+    gnet_inetaddr_new_async (hostname, port,  
+			     gnet_tcp_socket_connect_inetaddr_cb, 
+			     state);
 
-  /* Note that gnet_inetaddr_new_async can fail immediately and call
-     our callback which will delete the state.  The users callback
-     would be called in the process. */
-
-  if (id == NULL)
-    return NULL;
-
-  state->inetaddr_id = id;
+  /* On failure, gnet_inetaddr_new_async() returns NULL.  It will not
+     call the callback before it returns. */
+  if (state->inetaddr_id == NULL)
+    {
+      g_free (state);
+      return NULL;
+    }
 
   return state;
 }
@@ -116,17 +114,24 @@ gnet_tcp_socket_connect_inetaddr_cb (GInetAddr* inetaddr,
 
   if (status == GINETADDR_ASYNC_STATUS_OK)
     {
-      state->ia = inetaddr;
+      gpointer tcp_id;
+
+      state->ia = gnet_inetaddr_clone (inetaddr);
 
       state->inetaddr_id = NULL;
-      state->tcp_id = gnet_tcp_socket_new_async(inetaddr, 
-						gnet_tcp_socket_connect_tcp_cb, 
-						state);
-      /* Note that this call may delete the state. */
+
+      tcp_id = gnet_tcp_socket_new_async (inetaddr, 
+					  gnet_tcp_socket_connect_tcp_cb, 
+					  state);
+      /* gnet_tcp_socket_new_async() may call the callback before it
+	 returns.  state may have been deleted. */
+      if (tcp_id)
+	state->tcp_id = tcp_id;
     }
   else
     {
-      (*state->func)(NULL, NULL, GTCP_SOCKET_CONNECT_ASYNC_STATUS_INETADDR_ERROR, 
+      (*state->func)(NULL, NULL, 
+		     GTCP_SOCKET_CONNECT_ASYNC_STATUS_INETADDR_ERROR, 
 		     state->data);
       g_free(state);
     }
@@ -141,11 +146,18 @@ gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket,
   GTcpSocketConnectState* state = (GTcpSocketConnectState*) data;
 
   if (status == GTCP_SOCKET_NEW_ASYNC_STATUS_OK)
-    (*state->func)(socket, state->ia, GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK, state->data);
+    {
+      (*state->func)(socket, state->ia, 
+		     GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK, state->data);
+    }
   else
-    (*state->func)(NULL, NULL, GTCP_SOCKET_CONNECT_ASYNC_STATUS_TCP_ERROR, state->data);
+    {
+      (*state->func)(NULL, NULL, 
+		     GTCP_SOCKET_CONNECT_ASYNC_STATUS_TCP_ERROR, state->data);
+      gnet_inetaddr_delete (state->ia);
+    }
 
-  g_free(state);
+  g_free (state);
 }
 
 
@@ -158,7 +170,7 @@ gnet_tcp_socket_connect_tcp_cb (GTcpSocket* socket,
  * 
  */
 void
-gnet_tcp_socket_connect_async_cancel(GTcpSocketConnectAsyncID id)
+gnet_tcp_socket_connect_async_cancel (GTcpSocketConnectAsyncID id)
 {
   GTcpSocketConnectState* state = (GTcpSocketConnectState*) id;
 
@@ -170,9 +182,11 @@ gnet_tcp_socket_connect_async_cancel(GTcpSocketConnectAsyncID id)
     }
   else if (state->tcp_id)
     {
-      gnet_inetaddr_delete(state->ia);
-      gnet_tcp_socket_new_async_cancel(state->tcp_id);
+      gnet_inetaddr_delete (state->ia);
+      gnet_tcp_socket_new_async_cancel (state->tcp_id);
     }
+  else
+    g_assert_not_reached();
 
   g_free (state);
 }
@@ -247,7 +261,7 @@ gnet_tcp_socket_new_direct (const GInetAddr* addr)
   rv = connect(sockfd, &s->sa, sizeof(s->sa));
   if (rv != 0)
     {
-      close (sockfd);
+      GNET_CLOSE_SOCKET(s->sockfd);
       g_free (s);
       return NULL;
     }
@@ -384,7 +398,8 @@ gnet_tcp_socket_new_async_direct (const GInetAddr* addr,
   state->func = func;
   state->data = data;
   state->flags = flags;
-  state->connect_watch = g_io_add_watch(GNET_SOCKET_IOCHANNEL_NEW(s->sockfd),
+  state->iochannel = gnet_private_iochannel_new(s->sockfd);
+  state->connect_watch = g_io_add_watch(state->iochannel,
 					GNET_ANY_IO_CONDITION,
 					gnet_tcp_socket_new_async_cb, 
 					state);
@@ -401,8 +416,10 @@ gnet_tcp_socket_new_async_cb (GIOChannel* iochannel,
   GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) data;
   gint error, len;
 
-  /* Remove the watch now in case we don't return immediately */
   g_source_remove (state->connect_watch);
+  state->connect_watch = 0;
+  g_io_channel_unref (state->iochannel);
+  state->iochannel = NULL;
 
   errno = 0;
   if (!((condition & G_IO_IN) || (condition & G_IO_OUT)))
@@ -450,8 +467,11 @@ gnet_tcp_socket_new_async_cancel (GTcpSocketNewAsyncID id)
 {
   GTcpSocketAsyncState* state = (GTcpSocketAsyncState*) id;
 
-  g_source_remove(state->connect_watch);
-  gnet_tcp_socket_delete(state->socket);
+  if (state->connect_watch)
+    g_source_remove(state->connect_watch);
+  if (state->iochannel)
+    g_io_channel_unref (state->iochannel);
+  gnet_tcp_socket_delete (state->socket);
   g_free (state);
 }
 
@@ -520,10 +540,11 @@ gnet_tcp_socket_new_async_direct (const GInetAddr* addr,
   state->data = data;
   state->socket->sockfd = sockfd;
 
-  state->connect_watch = g_io_add_watch(GNET_SOCKET_IOCHANNEL_NEW(s->sockfd),
-					G_IO_IN | G_IO_ERR,
-					gnet_tcp_socket_new_async_cb, 
-					state);
+  state->connect_watch = 
+    g_io_add_watch(gnet_private_iochannel_new(s->sockfd),
+		   G_IO_IN | G_IO_ERR,
+		   gnet_tcp_socket_new_async_cb, 
+		   state);
 
 
   if (state->connect_watch <= 0)
@@ -663,7 +684,7 @@ gnet_tcp_socket_get_iochannel(GTcpSocket* socket)
   g_return_val_if_fail (socket != NULL, NULL);
 
   if (socket->iochannel == NULL)
-    socket->iochannel = GNET_SOCKET_IOCHANNEL_NEW(socket->sockfd);
+    socket->iochannel = gnet_private_iochannel_new(socket->sockfd);
    
   g_io_channel_ref (socket->iochannel);
 
@@ -1149,7 +1170,10 @@ gnet_tcp_socket_server_accept_async (GTcpSocket* socket,
   g_return_if_fail (!socket->accept_func);
 
   if (gnet_socks_get_enabled())
-    return gnet_private_socks_tcp_socket_server_accept_async (socket, accept_func, user_data);
+    {
+      gnet_private_socks_tcp_socket_server_accept_async (socket, accept_func, user_data);
+      return;
+    }
 
   /* Save callback */
   socket->accept_func = accept_func;
